/*===========================================================================
 *
 * File:	Dfpacked.CPP
 * Author:	Dave Humphrey (uesp@m0use.net)
 * Created On:	January 17, 2003
 *
 * Description
 *
 *=========================================================================*/

	/* Include Files */
#include "dfpacked.h"
#include "tcBitstream.h"
#include "tcDecoder.h"
#include <string.h>
#include <assert.h>
#include <stdlib.h>

/*===========================================================================
 *
 * Begin Local Definitions
 *
 *=========================================================================*/

	/* Callback data */
  PDFPACKCALLBACK l_pPackCallback = NULL;
  void*		  l_pCallbackData = NULL;

	/* Infomation about the extraction process */
  dfpackinfo_t l_PackInfo;

/*===========================================================================
 *		End of Local Definitions
 *=========================================================================*/


/*===========================================================================
 *
 * Function - void l_DFPackSetError (pErrorMsg);
 *
 * Sets the DFPack error message,
 *
 *=========================================================================*/
void l_DFPackSetError (const char* pErrorMsg) {
  ErrorHandler.AddError(ERR_CUSTOM, pErrorMsg);
 }
/*===========================================================================
 *		End of Function l_DFPackSetError()
 *=========================================================================*/


/*===========================================================================
 *
 * Function - int l_DFPackCallback (void);
 *
 * Calls the current packed callback function, if any.
 *
 *=========================================================================*/
int l_DFPackCallback (void) {
  if (l_pPackCallback != NULL) return l_pPackCallback(l_PackInfo, l_pCallbackData);
  return (0);
 }
/*===========================================================================
 *		End of Function l_DFPackCallback()
 *=========================================================================*/


/*===========================================================================
 *
 * Function - FILE* l_DFPackOpen (pFilename, pMode);
 *
 * Attempt to open the given filename file in the given mode.  Returns NULL
 * on any error.
 *
 *=========================================================================*/
FILE* l_DFPackOpen (const char* pFilename, const char* pMode) {
  FILE* pFileHandle;

	/* Attempt to open input PACKED.DAT file for reading */
  try {
    pFileHandle = fopen(pFilename, pMode);
   }
  catch(...) {
    l_DFPackSetError("Exception opening the file!");
    return (NULL);
  }

  l_DFPackSetError("Failed to open the file!");
  return (pFileHandle);
 }
/*===========================================================================
 *		End of Function l_DFPackOpen()
 *=========================================================================*/


/*===========================================================================
 *
 * Function - bool l_DFPackReadHeader (pFileHandle, Header);
 *
 * Local function to input the header data from the current position in 
 * the given file.  Returns false on any error.
 *
 *=========================================================================*/
bool l_DFPackReadHeader (FILE* pFileHandle, dfpackheader_t& Header) {
  int Result;
  
  Result = fread(&Header, 1, sizeof(dfpackheader_t), pFileHandle);

  if (Result != sizeof(dfpackheader_t)) {
    l_DFPackSetError("Read error, failed to input the header data!");
    return (false);
   }

  return (true);
 }
/*===========================================================================
 *		End of Function l_DFPackReadHeader()
 *=========================================================================*/


/*===========================================================================
 *
 * Function - bool l_DFPackReadRecords (pFileHandle, pRecords, Header);
 *
 * Inputs the record data from the file.  Returns false on any error.
 *
 *=========================================================================*/
bool l_DFPackReadRecords (FILE* pFileHandle, dfpackrecord_t* pRecords, const dfpackheader_t& Header) {
  int Result;

		/* Jump to the correct point in the file */
  Result = fseek(pFileHandle, Header.OffsetRecordDir, SEEK_SET);

  if (Result != 0) {
    l_DFPackSetError("Seek error, failed to find the record data!");
    return (false);
    }

	/* Input the record data all at once */
  Result = fread(pRecords, 2, sizeof(dfpackrecord_t), pFileHandle);

  if (Result != sizeof(dfpackrecord_t)) {
    l_DFPackSetError("Read error, failed to input the record data!");
    return (false);
    }

  return (true);
 }
/*===========================================================================
 *		End of Function l_DFPackReadRecords()
 *=========================================================================*/


/*===========================================================================
 *
 * Function - bool l_DFPackUncompress (pInputFile, pOutputFile, Record);
 *
 * Uncompresses the given file record from the current position in the
 * input file, saving the results in the output file.  Returns false
 * on any error.
 *
 *=========================================================================*/
bool l_DFPackUncompress (FILE* pInputFile, FILE* pOutputFile, const dfpackrecord_t& Record) {
  char*		pBuffer;
  long		nLength = 0;
  long		nStreamCount = 0;
  unsigned int	nStreamLength = 0;
  unsigned int	Result;
  int		CBResult;
  bool		ReturnResult = true;
  tcBitstream*  pBitstream = NULL;
  tcDecoder*    pDecoder   = NULL;

	/* Create a memory buffer */
  pBuffer = new char[DFPACK_BUFFER_SIZE];

  if (pBuffer == NULL) {
    l_DFPackSetError("Critical memory error!");
    return (false);
   }

	/* Handle all streams until source file complete */
  while ( nLength < Record.FileLength ) {

  		/* Initialise Bitstream and Decoder objects */
    pBitstream = new tcBitstream( pInputFile );
    pDecoder = new tcDecoder( pBitstream );
    assert(pBitstream != NULL);
    assert(pDecoder   != NULL);

		/* Invoke decode method */
    pDecoder->Decode( pBuffer, &nStreamLength, DFPACK_BUFFER_SIZE );

		/* Seek past header to next stream */
    Result = fseek( pInputFile, DFPACK_RECORDHEADER_LENGTH, SEEK_CUR );

    if (Result != 0) {
      ReturnResult = false;
      l_DFPackSetError("Seek error, failed to find the compressed data!");
      break;
     }

		/* Append decompressed stream to output file */
    Result = fwrite( pBuffer, 1, nStreamLength, pOutputFile );
    
    if (Result != nStreamLength) {
      ReturnResult = false;
      l_DFPackSetError("Write error, failed to output the file data!");
      break;
     }

		/* Clear current Bitstream and Decoder objects */
    delete pBitstream;
    delete pDecoder;
    pBitstream = NULL;
    pDecoder   = NULL;

		/* Increment decoded length */
    nLength += nStreamLength;

		/* Increment stream count */
    nStreamCount++;

	/* Update the callback function */
    l_PackInfo.NumBytes += nStreamLength;
    CBResult = l_DFPackCallback();

    if (CBResult < 0) {
      ReturnResult = false;
      break;
     }
   }

	/* Delete objects in case of error */
  if (pBitstream != NULL) delete pBitstream;
  if (pDecoder   != NULL) delete pDecoder;
  delete[] pBuffer;

  return (ReturnResult);
 }
/*===========================================================================
 *		End of Function l_DFPackUncompress()
 *=========================================================================*/


/*===========================================================================
 *
 * Function - bool l_DFPackExtractRecord (pFileHandle, pOutputPath, Record);
 *
 * Attempts to extract the given record from the current PACKED.DAT file,
 * saving the new file to the given path (plus an relative path).  Returns
 * false on any error.
 *
 *=========================================================================*/
bool l_DFPackExtractRecord (FILE* pFileHandle, const char* pOutputPath, const dfpackrecord_t& Record) {
  char  OutputFilename[_MAX_PATH+1];
  FILE* pOutputHandle;
  int   Result;
  bool  fResult;

  	/* Seek to the start of the compressed file stream, skipping header */
  Result = fseek(pFileHandle, Record.Offset + DFPACK_RECORDHEADER_LENGTH, SEEK_SET);

  if (Result != 0) {
    l_DFPackSetError("Seek error, failed to find the compressed data!");
    return (false);
   }

	/* Initialize the callback info */
  l_PackInfo.FileSize  = Record.FileLength;
  l_PackInfo.NumBytes  = 0;
  l_PackInfo.pFilename = Record.Filename;
  l_DFPackCallback();

	/* Compose filename and attempt to open the output */
  _makepath(OutputFilename, NULL, pOutputPath, Record.Filename, NULL);
  pOutputHandle = l_DFPackOpen(OutputFilename, "wb");
  if (pOutputHandle == NULL) return (false);

	/* Uncompress the file */
  fResult = l_DFPackUncompress (pFileHandle, pOutputHandle, Record);
  fclose (pOutputHandle);
  return (fResult);
 }
/*===========================================================================
 *		End of Function l_DFPackExtractRecord()
 *=========================================================================*/


/*===========================================================================
 *
 * Function - bool DFPackExtract (pInputFile, pOutputPath, pCallback, pData);
 *
 * Attempts to extract the given PACKED.DAT file to the given path.
 * returns false on any error.  Calls the given function, if any, during
 * the extraction process.
 *
 *=========================================================================*/
bool DFPackExtract (const char* pInputFile, const char* pOutputPath, 
			PDFPACKCALLBACK pCallback, void *pData) {
  dfpackheader_t Header;
  dfpackrecord_t Records[DFPACK_NUM_RECORDS];
  FILE*		 pFileHandle;
  bool		 Result;

	/* Save the callback data */
  l_pPackCallback = pCallback;
  l_pCallbackData = pData;
  
	/* Attempt to open file for input */
  pFileHandle = l_DFPackOpen(pInputFile, "rb");
  if (pFileHandle == NULL) return (false);

	/* Attempt to input the packed.dat header data */
  Result = l_DFPackReadHeader(pFileHandle, Header);
  if (Result) Result = l_DFPackReadRecords(pFileHandle, &Records[0], Header);

	/* Extract the files */
  if (Result) Result = l_DFPackExtractRecord (pFileHandle, pOutputPath, Records[0]);
  if (Result) Result = l_DFPackExtractRecord (pFileHandle, pOutputPath, Records[1]);

  fclose(pFileHandle);
  return (Result);
 }
/*===========================================================================
 *		End of Function DFPackExtract()
 *=========================================================================*/


