/*===========================================================================
 *
 * File:	Mwesmfile.CPP
 * Author:	Dave Humphrey (uesp@m0use.net)
 * Created On:	Wednesday, September 04, 2002
 *
 * Description
 *
 *=========================================================================*/

	/* Include Files */
#include "mwesmfile.h"
#include "mwesm_tes3.h"
#include "mwesm_weap.h"
#include "mwesm_armo.h"
#include "mwesm_clot.h"
#include "mwesm_repa.h"
#include "mwesm_appa.h"
#include "mwesm_misc.h"
#include "mwesm_lock.h"
#include "mwesm_prob.h"
#include "mwesm_ingr.h"
#include "mwesm_book.h"
#include "mwesm_alch.h"


/*===========================================================================
 *
 * Begin Local Definitions
 *
 *=========================================================================*/
  DEFINE_FILE("MWESMFile.cpp");
/*===========================================================================
 *		End of Local Definitions
 *=========================================================================*/


/*===========================================================================
 *
 * Begin ESM Base Record Definitions
 *
 *=========================================================================*/

  mwesmbaserecdef_t l_RecordDefs[] = {
	MWESM_RECORD_DEF("WEAP", CMwEsmRecordWeap::Create)
	MWESM_RECORD_DEF("ARMO", CMwEsmRecordArmo::Create)
	MWESM_RECORD_DEF("CLOT", CMwEsmRecordClot::Create)
	MWESM_RECORD_DEF("REPA", CMwEsmRecordRepa::Create)
	MWESM_RECORD_DEF("APPA", CMwEsmRecordAppa::Create)
	MWESM_RECORD_DEF("MISC", CMwEsmRecordMisc::Create)
	MWESM_RECORD_DEF("LOCK", CMwEsmRecordLock::Create)
	MWESM_RECORD_DEF("PROB", CMwEsmRecordProb::Create)
	MWESM_RECORD_DEF("INGR", CMwEsmRecordIngr::Create)
	MWESM_RECORD_DEF("BOOK", CMwEsmRecordBook::Create)
	MWESM_RECORD_DEF("ALCH", CMwEsmRecordAlch::Create)
	MWESM_RECORD_DEF("    ", NULL)		/* Must be last record */
   };
/*===========================================================================
 *		End of ESM Base Record Definitions
 *=========================================================================*/



/*===========================================================================
 *
 * Function - int l_FindEsmRecordDef (Header);
 *
 * Local function which attempts to find a record definition in the local
 * array with a matching header name.  Returns -1 if the record was not
 * found.
 *
 *=========================================================================*/
int l_FindEsmRecordDef (const mwesmbaseheader_t& Header) {
  bool Result;
  int  DefIndex;

	/* Search through the entire record def array */
  for (DefIndex = 0; l_RecordDefs[DefIndex].pfnCreate != NULL; DefIndex++) {
    Result = MwEsmCompareHeaderName(Header.Name, l_RecordDefs[DefIndex].Name);
    if (Result) return (DefIndex);
   }

  return (-1);
 }
/*===========================================================================
 *		End of Function l_FindEsmRecordDef()
 *=========================================================================*/


/*===========================================================================
 *
 * Function - CMwEsmBaseRecord* l_CreateBaseRecord (Header);
 *
 * Local function which attempts to create and return a new pointer for a class
 * with the given header. If a matching class does not exist, the base
 * record class is used.  Stores the header information to the class.
 *
 *=========================================================================*/
CMwEsmBaseRecord* l_CreateBaseRecord (const mwesmbaseheader_t& Header) {
  DEFINE_FUNCTION("l_CreateBaseRecord()");
  CMwEsmBaseRecord* pNewRecord;
  int		    DefIndex;

	/* Search for a matching record definition */
  DefIndex = l_FindEsmRecordDef(Header);

	/* Use base record class if no matching class found */
  if (DefIndex < 0) {
    CreatePointer(pNewRecord, CMwEsmBaseRecord);
   }
  else {
    pNewRecord = l_RecordDefs[DefIndex].pfnCreate();
   }

  pNewRecord->SetHeader(Header);
  return (pNewRecord);
 }
/*===========================================================================
 *		End of Function l_FindEsmRecordDef()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CMwEsmFile Constructor
 *
 *=========================================================================*/
CMwEsmFile::CMwEsmFile () {
  m_ppRecords      = NULL;
  m_fnLoadCallback = NULL;

  m_NumRecords       = 0;
  m_AllocatedRecords = 0;

  m_SkipData     = false;
  m_SkipBaseData = false;

  ResizeRecordsArray(MWESM_DEFAULT_NUMRECORDS);
 }
/*===========================================================================
 *		End of Class CMwEsmFile Constructor
 *=========================================================================*/


/*===========================================================================
 *
 * Class CMwEsmFile Method - void Destroy (void);
 *
 * Deletes the contents of the file.
 *
 *=========================================================================*/
void CMwEsmFile::Destroy (void) {
  DEFINE_FUNCTION("CMwEsmFile::Destroy()");

  DeleteRecords();

	/* Delete the array of allocated record pointers */
  DestroyArrayPointer(m_ppRecords);

  m_NumRecords       = 0;
  m_AllocatedRecords = 0;
 }
/*===========================================================================
 *		End of Class Method CMwEsmFile::Destroy()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CMwEsmFile Method - void AddRecord (pNewRecord);
 *
 * Protected class method which adds the given record to the end of the
 * record array.  The array is resized if required.
 *
 *=========================================================================*/
void CMwEsmFile::AddRecord (CMwEsmBaseRecord* pNewRecord) {

	/* Resize the array if required */
  if (m_NumRecords >= m_AllocatedRecords) ResizeRecordsArray(m_AllocatedRecords*2);

	/* Add the record to the end of the array */
  m_ppRecords[m_NumRecords] = pNewRecord;
  m_NumRecords++;
 }
/*===========================================================================
 *		End of Class Method CMwEsmFile::AddRecord()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CMwEsmFile Method - void DeleteRecords (void);
 *
 * Protected class method to delete all allocated records.
 *
 *=========================================================================*/
void CMwEsmFile::DeleteRecords (void) {
  DEFINE_FUNCTION("CMwEsmFile::DeleteRecords()");
  int RecordIndex;

	/* Ignore if array has not yet been initialized */
  if (m_ppRecords == NULL) return;

	/* Delete all allocated record pointers */
  for (RecordIndex = 0; RecordIndex < m_NumRecords; RecordIndex++) {
    DestroyPointer(m_ppRecords[RecordIndex]);
   }

  m_NumRecords = 0;
 }
/*===========================================================================
 *		End of Class Method CMwEsmFile::DeleteRecords()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CMwEsmFile Method - int ForEachRecord (fnCallback, pUserData);
 *
 * Uses the given callback function to call with each currently defined
 * record in the array.  Returns a negative value on any error.
 *
 *=========================================================================*/
int CMwEsmFile::ForEachRecord (MWESM_FOREACHCALLBACK fnCallback, void* pUserData) {
  int Result = 0;
  int RecordIndex;
	
	/* Ensure valid input */
  if (fnCallback == NULL) return (-1);

	/* Call the function for each defined record in array */
  for (RecordIndex = 0; RecordIndex < m_NumRecords; RecordIndex++) {
    Result = fnCallback(m_ppRecords[RecordIndex], pUserData);
    if (Result < 0) break;
   }

  return (Result);
 } 
/*===========================================================================
 *		End of Class Method CMwEsmFile::ForEachRecord()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CMwEsmFile Method - bool LoadESM (pFilename);
 *
 * Attempt to load the ESM file specified by the input filename.  Returns
 * false on any error.
 *
 *=========================================================================*/
bool CMwEsmFile::LoadESM (const char* pFilename) {
  DEFINE_FUNCTION("CMwEsmFile::LoadESM()");
  bool		    Result;
  int		    iResult;
  float		    PercentDone;
  filepos_t	    FileSize;
  mwesmbaseheader_t RecordHeader;
  CMwEsmBaseRecord* pNewRecord = NULL;

  
	/* Attempt to open file for binary input */
  Result = Open(pFilename, "rb");
  if (!Result) return (false);

	/* Ensure all the current records are deleted */
  DeleteRecords();
  FileSize = GetFileSize();

	/* Attempt to read entire file */
  while (!IsEOF()) {

		/* Determine if there is anything left to read, preventing an error */
    if (Tell() >= FileSize) break; 

		/* Peek at the next record header */
    Result = CMwEsmBaseRecord::ReadRecordHeader(*this, RecordHeader);
    if (!Result) break;

		/* Call the user's callback function, if any */
    if (m_fnLoadCallback != NULL) {
      PercentDone = RecordHeader.StartOffset / (float) FileSize * 100.0f;
      iResult = m_fnLoadCallback(RecordHeader, PercentDone);

		/* Abort the loading process */
      if (iResult < 0) {
        Result = false;
	break;
       }
     }

		/* Create a new record and attempt to read it */
    pNewRecord = l_CreateBaseRecord(RecordHeader);

    if (pNewRecord->IsBaseRecord())
      pNewRecord->SetSkipData(m_SkipBaseData);
    else
      pNewRecord->SetSkipData(m_SkipData);

    Result = pNewRecord->Read(*this);
    if (!Result) break;

		/* On success, add record to array */
    AddRecord(pNewRecord);
    pNewRecord = NULL;
   }

  Close();
  DestroyPointer(pNewRecord);
  return (Result);
 }
/*===========================================================================
 *		End of Class Method CMwEsmFile::LoadESM()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CMwEsmFile Method - bool LoadESM (pFilename, CallbackFunc);
 *
 * Loads the given ESM file using the given user callback function to update
 * the user on the load progress.  Returns false on any error.
 *
 *=========================================================================*/
bool CMwEsmFile::LoadESM (const char* pFilename, MWESM_LOADCALLBACK CallbackFunc) {
  bool Result;

  m_fnLoadCallback = CallbackFunc;
  Result = LoadESM(pFilename);
  m_fnLoadCallback = NULL;

  return (Result);
 }
/*===========================================================================
 *		End of Class Method CMwEsmFile::LoadESM()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CMwEsmFile Method - bool OutputItems (File);
 *
 *=========================================================================*/
bool CMwEsmFile::OutputItems (CGenFile& File) {
  bool Result;
  int  RecordIndex;
	
	/* Output header */
  File.Printf ("ID, Name, Type, Weight (kg), Value\n");
  
	/* Attempt to output all records */
  for (RecordIndex = 0; RecordIndex < m_NumRecords; RecordIndex++) {
    Result = m_ppRecords[RecordIndex]->OutputItem(File);
    if (!Result) return (false);
   }

  return (true);
 }
/*===========================================================================
 *		End of Class Method CMwEsmFile::OutputItems()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CMwEsmFile Method - bool ResizeRecordsArray (NewSize);
 *
 * Protected class method which resizes the array of records. The current
 * record pointers are copied into the new array.
 *
 *=========================================================================*/
bool CMwEsmFile::ResizeRecordsArray (const int NewSize) {
  DEFINE_FUNCTION("CMwEsmFile::ResizeRecordsArray()");
  CMwEsmBaseRecord** ppNewArray;

	/* Ensure valid input */
  if (NewSize <= 0 || NewSize < m_NumRecords) return (false);

	/* Create the new array of pointers */
  CreateArrayPointer(ppNewArray, CMwEsmBaseRecord*, NewSize);

	/* Copy over the existing pointer array, if any, and delete old array */
  if (m_ppRecords != NULL) {
    memcpy(ppNewArray, m_ppRecords, sizeof(CMwEsmBaseRecord*)*m_NumRecords);
    DestroyArrayPointer(m_ppRecords);
   }

  m_ppRecords = ppNewArray;
  m_AllocatedRecords = NewSize;
  return (true);
 }
/*===========================================================================
 *		End of Class Method CMwEsmFile::ResizeRecordsArray()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CMwEsmFile Method - bool SaveESM (pFilename);
 *
 * Attempt to save the ESM file specified by the output filename.  Returns
 * false on any error.
 *
 *=========================================================================*/
bool CMwEsmFile::SaveESM (const char* pFilename) {
  DEFINE_FUNCTION("CMwEsmFile::SaveESM()");
  bool Result;
  int  RecordIndex;

	/* Attempt to open file for binary output */
  Result = Open(pFilename, "wb");
  if (!Result) return (false);

	/* Attempt to output the entire file */
  for (RecordIndex = 0; RecordIndex < m_NumRecords; RecordIndex++) {
    Result = m_ppRecords[RecordIndex]->Write(*this);
    if (!Result) break;
   }

  Close();
  return (Result);
 }
/*===========================================================================
 *		End of Class Method CMwEsmFile::SaveESM()
 *=========================================================================*/

