/*===========================================================================
 *
 * File:	Mwesm_baserec.H
 * Author:	Dave Humphrey (uesp@m0use.net)
 * Created On:	Wednesday, September 04, 2002
 *
 * Defines the CMwEsmBaseRecord class from which all Morrowind ESM top
 * level records are defined.
 *
 *=========================================================================*/
#ifndef __MWESM_BASEREC_H
#define __MWESM_BASEREC_H


/*===========================================================================
 *
 * Begin Required Includes
 *
 *=========================================================================*/
  #include "dl_err.h"
  #include "genfile.h"
/*===========================================================================
 *		End of Required Includes
 *=========================================================================*/


/*===========================================================================
 *
 * Begin Definitions
 *
 *=========================================================================*/
  
	/* Size, in bytes, of the header name */
  #define MWESM_HEADERNAME_SIZE (4)

	/* Size, in bytes, of the base record headers */
  #define MWESM_BASEHEADER_SIZE (16)

	/* Macro for registering an unknown subrecord */
  #define MWESM_BADSUBRECORD(Name, File) ErrorHandler.AddError(ERR_CUSTOM, "Unknown ESM subrecord '%4.4s' (Offset 0x%08lX)!", Name, File.Tell())

/*===========================================================================
 *		End of Definitions
 *=========================================================================*/


/*===========================================================================
 *
 * Begin Type Definitions
 *
 *=========================================================================*/
#pragma pack(push, 1)

	/* Header for base records */
  typedef struct {
    char Name[MWESM_HEADERNAME_SIZE];	/* Header identifier */
    long Size;				/* Size of record, in bytes, not including header */
    char Unknown[8];			/* Unknown data, NULL? */

    filepos_t StartOffset;		/* Start offset of header in file (added)*/
   } mwesmbaseheader_t;

	/* Method definition type for creating new objects */
  class CMwEsmBaseRecord;
  typedef CMwEsmBaseRecord* (*MWESMCREATE_FUNC) (void);

#pragma pack(pop)
/*===========================================================================
 *		End of Type Definitions
 *=========================================================================*/


/*===========================================================================
 *
 * Begin Class CMwEsmBaseRecord Definition
 *
 * The base class defining a generic top level record in Morrowind's ESM
 * plugin files.  The data is stored by default in  a single allocated pointer.
 *
 *=========================================================================*/
class CMwEsmBaseRecord {

  /*---------- Begin Protected Class Members --------------------*/
protected:
  mwesmbaseheader_t	m_Header;	/* Describes the record */
  bool			m_HeaderRead;	/* Has header been input? */

  bool			m_SkipData;	/* Do not input record data */
  bool			m_DataRead;	/* Has the data been input? */
  char*			m_pData;	/* The raw data, if loaded */


  /*---------- Begin Protected Class Methods --------------------*/
protected:

	/* Helper input/output functions */
  virtual bool ReadHeader  (CGenFile& File);
  virtual bool WriteHeader (CGenFile& File);
  virtual bool SkipData    (CGenFile& File);
  virtual bool ReadData    (CGenFile& File);
  virtual bool WriteData   (CGenFile& File);


  /*---------- Begin Public Class Methods -----------------------*/
public:

	/* Class Constructors/Destructors */
  CMwEsmBaseRecord();
  virtual ~CMwEsmBaseRecord() { Destroy(); }
  virtual void Destroy (void);

	/* Returns a new pointer of itself */
  static CMwEsmBaseRecord* Create (void);

	/* Get record class members */
  const char*  GetName		(void) const { return (m_Header.Name); }
  virtual const char*  GetTypeString    (void) const { return ("Base"); }
  long         GetSize		(void) const { return (m_Header.Size); }
  long	       GetEndOffset	(void) const { return (m_Header.StartOffset + m_Header.Size + 16); }
  virtual bool IsBaseRecord	(void) const { return (true); }

	/* Output the basic item stats to the given file */
  virtual bool OutputItem (CGenFile& File) { return (true); }

	/* Static method to allow input of record header */
  static bool ReadRecordHeader (CGenFile& File, mwesmbaseheader_t& Header);

	/* Input record from current position in file */
  virtual bool Read (CGenFile& File);

	/* Set class members */
  virtual void SetSkipData (const bool Flag) { m_SkipData     = Flag; } 
  virtual void SetHeader   (const mwesmbaseheader_t NewHeader) { m_Header = NewHeader; m_HeaderRead = true; }

	/* Output record to current position in file */
  virtual bool Write (CGenFile& File);


 };
/*===========================================================================
 *		End of Class CMwEsmBaseRecord Definition
 *=========================================================================*/


/*===========================================================================
 *
 * Begin Function Prototypes
 *
 *=========================================================================*/

	/* Compares two header names for equality */
  bool MwEsmCompareHeaderName (const char* pHeader1, const char* pHeader2);

	/* Input a subrecord name/size pair */
  bool MwEsmReadSubrecordName (CGenFile& File, char* pName, long& StringSize);

/*===========================================================================
 *		End of Function Prototypes
 *=========================================================================*/




#endif
/*===========================================================================
 *		End of File Mwesm_baserec.H
 *=========================================================================*/
