/*===========================================================================
 *
 * File:	Gengraphdata.CPP
 * Author:	Dave Humphrey (uesp@m0use.net)
 * Created On:	Thursday, October 17, 2002
 *
 * Description
 *
 * 30 October 2003
 *	- Added the buffering of data so the display can be properly updated
 *	  when autofitting during a scan in TCScan.
 *
 * 10 November 2003
 *	- Updated scan drawing methods to properly update the graph if the
 *	  scan is going backwards as well as forwards.
 *
 * 29 January 2004
 *	- Proper clipping is used now when drawing a regular graph.
 *	- Added the capability of displaying data points in a regular graph.
 *
 *=========================================================================*/

	/* Include Files */
#include "stdafx.h"
#include "gengdata.h"


/*===========================================================================
 *
 * Begin Local Definitions
 *
 *=========================================================================*/
  DEFINE_FILE("GenGData.cpp");
/*===========================================================================
 *		End of Local Definitions
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Constructor
 *
 *=========================================================================*/
CGenGraphData::CGenGraphData () {
  DEFINE_FUNCTION("CGenGraphData::CGenGraphData()");

  m_pXAxis = NULL;
  m_pYAxis = NULL;
  
  m_XData.m_Referenced = false;
  m_XData.m_NumPoints  = 0;
  m_XData.m_SkipBytes  = 0;
  m_XData.m_pData      = NULL;

  m_YData.m_Referenced = false;
  m_YData.m_NumPoints  = 0;
  m_YData.m_SkipBytes  = 0;
  m_YData.m_pData      = NULL;

  m_pFramePen = new CPen(PS_SOLID, 1, GENGRAPH_COLOR_BLACK);
  m_pBGBrush  = new CBrush(GENGRAPH_COLOR_WHITE);
  m_DataColor = GENGRAPH_COLOR_RED;
  m_pDataPen  = new CPen(PS_SOLID, 1, m_DataColor);
  ValidateNewPointer(m_pFramePen);
  ValidateNewPointer(m_pBGBrush);
  ValidateNewPointer(m_pDataPen);

  m_DrawBG = true;

  m_Type = GENGRAPH_TYPE_DEFAULT;

  m_ScanColor = GENGRAPH_COLOR_BLUE;
  m_pScanPen  = new CPen(PS_SOLID, 1, m_ScanColor);
  ValidateNewPointer(m_pScanPen);
  m_HasLastPoint     = false;
  m_AutoScroll       = true;
  m_DisplayScan      = true;
  m_ForceCleanScan   = false;
  m_ScanForwards     = true;
  m_LastScanForwards = true;
  m_NormalizeXAxis   = false;
  m_DisplayPoints    = false;
  m_ScanX = 0;
  m_LastX = 0;
  m_LastY = 0;

  m_NumBuffers       = 0;
  m_BufferStartIndex = 0;
 }
/*===========================================================================
 *		End of Class CGenGraphData Constructor
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Destructor
 *
 *=========================================================================*/
CGenGraphData::~CGenGraphData () {
  DEFINE_FUNCTION("CGenGraphData::~CGenGraphData()");

  Destroy();

  DestroyPointer(m_pFramePen);
  DestroyPointer(m_pBGBrush);
  DestroyPointer(m_pDataPen);
  DestroyPointer(m_pScanPen);
 }
/*===========================================================================
 *		End of Class CGenGraphData Destructor
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void Destroy (void);
 *
 *=========================================================================*/
void CGenGraphData::Destroy (void) {
  //DEFINE_FUNCTION("CGenGraphData::Destroy()");
  m_pXAxis = NULL;
  m_pYAxis = NULL;

  m_NumBuffers       = 0;
  m_BufferStartIndex = 0;

  DestroyData(); 
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::Destroy()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void AddBufferPoint (X, Y);
 *
 * Adds the given point to the circular data buffer for scan graphs.
 *
 *=========================================================================*/
void CGenGraphData::AddBufferPoint (const float X, const float Y) {

	/* Store the data point */
  m_XDataBuffer[m_BufferStartIndex] = X;
  m_YDataBuffer[m_BufferStartIndex] = Y;

	/* Increment the ring buffer counters as required */
  if (m_NumBuffers < GENGDATA_BUFFER_SIZE) m_NumBuffers++;
  m_BufferStartIndex = (m_BufferStartIndex + 1 ) % GENGDATA_BUFFER_SIZE;
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::AddBufferPoint()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void AutoScrollXAxis (X, pDC);
 *
 * For a scan graph, checks to see if the X-Axis needs to be scrolled.
 * Protected class method.
 *
 *=========================================================================*/
void CGenGraphData::AutoScrollXAxis (const float X, CDC* pDC) {
  float Range;

	/* Ignore if auto scrolling is disabled */
  if (!m_AutoScroll || !m_DisplayScan) return;

	/* Scroll X-Axis forwards if required */
  if (X > m_pXAxis->GetEndCoor()) {
    m_ForceCleanScan = true;
    ClearScan(X, pDC);
    m_ForceCleanScan = false;

      		/* Reset the graph start/end coordinates and redraw */
    Range = m_pXAxis->GetEndCoor() - m_pXAxis->GetStartCoor();
    m_pXAxis->SetEndCoor(m_pXAxis->GetEndCoor() + Range);
    m_pXAxis->SetStartCoor(m_pXAxis->GetStartCoor() + Range);
    m_pXAxis->OnPaint(pDC);
   }
	/* Scroll X-Axis backwards if required */
  else if (X < m_pXAxis->GetStartCoor()) {
    m_ForceCleanScan = true;
    ClearScan(X, pDC);
    m_ForceCleanScan = false;

      		/* Reset the graph start/end coordinates and redraw */
    Range = m_pXAxis->GetEndCoor() - m_pXAxis->GetStartCoor();
    m_pXAxis->SetEndCoor(m_pXAxis->GetEndCoor() - Range);
    m_pXAxis->SetStartCoor(m_pXAxis->GetStartCoor() - Range);
    m_pXAxis->OnPaint(pDC);
   }

 }
/*===========================================================================
 *		End of Class Method CGenGraphData::AutoScrollXAxis()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void DestroyData (void);
 *
 *=========================================================================*/
void CGenGraphData::DestroyData (void) {
  DEFINE_FUNCTION("CGenGraphData::DestroyData()");

  if (!m_XData.m_Referenced) {
    DestroyArrayPointer(m_XData.m_pData);
   }

  if (!m_YData.m_Referenced) {
    DestroyArrayPointer(m_YData.m_pData);
   }

  m_XData.m_Referenced = false;
  m_XData.m_NumPoints  = 0;
  m_XData.m_SkipBytes  = 0;
  m_XData.m_pData      = NULL;

  m_YData.m_Referenced = false;
  m_YData.m_NumPoints  = 0;
  m_YData.m_SkipBytes  = 0;
  m_YData.m_pData      = NULL;

  m_HasLastPoint = false;
  m_LastX = 0;
  m_LastY = 0;
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::DestroyData()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void ClearScan (const float X, CDC* pDC);
 *
 * Used when drawing a scan graph to clear the current scan line.
 *
 *=========================================================================*/
void CGenGraphData::ClearScan (const float X, CDC* pDC) {
  //DEFINE_FUNCTION("CGenGraphData::ClearScan()");
  int  LogX1;
  int  LogX2;

	/* Ignore if required */
  if (!m_DisplayScan) return;

	/* Clear from the previous scan line to the current point */
  LogX1 = GetScanLocation(m_ScanX, m_LastScanForwards);
  LogX2 = GetScanLocation(X,       m_ScanForwards);

	/* Ignore if scan line hasn't moved */
  if (!m_ForceCleanScan && LogX1 == LogX2) return;

	/* Clear the scan area */
  pDC->SelectStockObject(NULL_PEN);  
  pDC->SelectObject(m_pBGBrush);

  if (m_ScanForwards) 
    pDC->Rectangle(LogX1, m_pYAxis->GetTop()+1, LogX2+2, m_pYAxis->GetBottom()+1);
  else
    pDC->Rectangle(LogX2, m_pYAxis->GetTop()+1, LogX1+2, m_pYAxis->GetBottom()+1);

  pDC->SelectStockObject(BLACK_PEN);
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::ClearScan()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - int DrawBG (pDC);
 *
 *=========================================================================*/
int CGenGraphData::DrawBG (CDC* pDC) { 

  pDC->SelectObject(m_pFramePen);
  pDC->SelectObject(m_pBGBrush);

	/* Draw the frame using the axis coordinates, if valid */
  if (HasValidAxes()) {

    if (m_DrawBG)
      pDC->Rectangle(m_pXAxis->GetLeft(),    m_pYAxis->GetTop(), 
		     m_pXAxis->GetRight()+1, m_pYAxis->GetBottom()+1);
    else {
      pDC->MoveTo(m_pXAxis->GetLeft(),    m_pYAxis->GetTop());
      pDC->LineTo(m_pXAxis->GetLeft(),    m_pYAxis->GetBottom());
      pDC->LineTo(m_pXAxis->GetRight(),   m_pYAxis->GetBottom());
      pDC->LineTo(m_pXAxis->GetRight(),   m_pYAxis->GetTop());
      pDC->LineTo(m_pXAxis->GetLeft(),    m_pYAxis->GetTop());
     }
   }

  return (0);
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::DrawBG()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - int DrawBuffer (pDC);
 *
 * Outputs the data in the current data buffer, if any.
 *
 *=========================================================================*/
int CGenGraphData::DrawBuffer (CDC* pDC) {
  int   Index;
  int   PointCount;
  int   StartIndex;
  bool  FirstPoint = true;
  float X;
  float Y;
  int   XPos;
  int   YPos;
	
	/* Find out the starting index in the circular buffer */
  StartIndex = m_BufferStartIndex;
  if (m_NumBuffers < GENGDATA_BUFFER_SIZE) StartIndex = 0;
  pDC->SelectObject(m_pDataPen);

	/* Output all points in the buffer */
  for (Index = StartIndex, PointCount = 0; PointCount < m_NumBuffers; 
		    Index = (Index + 1 ) % GENGDATA_BUFFER_SIZE,
		    PointCount++) {
    X = m_XDataBuffer[Index];
    Y = m_YDataBuffer[Index];

		/* Check X-data limits */
    if (X < m_pXAxis->GetStartCoor()) continue;
    if (X > m_pXAxis->GetEndCoor())   break;
    XPos = m_pXAxis->LimitCoorToLogical(X);
    YPos = m_pYAxis->LimitCoorToLogical(Y);

    if (FirstPoint) {
      pDC->MoveTo(XPos, YPos);
      FirstPoint = false;
     }
    else {
      pDC->LineTo(XPos, YPos);
     }
   }

	/* Reset so that the next point drawn is drawn correctly */
  m_HasLastPoint = false;
  return (0);
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::DrawBuffer()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - int DrawData (pDC);
 *
 *=========================================================================*/
int CGenGraphData::DrawData (CDC* pDC) {
  int		Index;
  int		LogX;
  int		LogY;
  graphdata_t	DataX;
  graphdata_t	DataY;
  bool		ValidLastPoint;
  float		XRange;
  CRgn		NewRegion;

	/* Ensure valid data and axes */
  if (!HasValidData() || !HasValidAxes()) return (-1);
  pDC->SelectObject(m_pDataPen);
  ValidLastPoint = false;
  XRange = m_pXAxis->GetEndCoor() - m_pXAxis->GetStartCoor();

	/* Create a clipping region around the graph to be drawn */
  NewRegion.CreateRectRgn(m_pXAxis->GetLeft(), m_pYAxis->GetTop(), m_pXAxis->GetRight(), m_pYAxis->GetBottom());
  pDC->SelectClipRgn(&NewRegion);

	/* Output all visible points */
  for (Index = 0; Index < m_YData.m_NumPoints; Index++) {

    if (m_XData.m_pData == NULL && XRange != 0) {
      if (m_NormalizeXAxis) {
        DataX = ((graphdata_t) Index / (graphdata_t) m_YData.m_NumPoints * XRange) + m_pXAxis->GetStartCoor();
       }
      else {
        DataX = (graphdata_t) Index;
       }
     }
    else if (Index >= m_XData.m_NumPoints)
      break;
    else if (m_XData.m_SkipBytes == 0)
      DataX = m_XData.m_pData[Index];
    else
      DataX = *(float *) (((byte *)m_XData.m_pData) + (4 + m_XData.m_SkipBytes) * Index);

    if (m_YData.m_SkipBytes == 0)
      DataY = m_YData.m_pData[Index];
    else
      DataY = *(float *) (((byte *)m_YData.m_pData) + (4 + m_YData.m_SkipBytes) * Index);

		/* Ensure the point is visible in the current graph */
  /*  if (!m_pXAxis->IsValidPoint(DataX) || !m_pYAxis->IsValidPoint(DataY)) {
      ValidLastPoint = false;
      continue;
     } //*/

		/* Convert coordinates to logical */
    LogX = m_pXAxis->FastCoorToLogical(DataX);
    LogY = m_pYAxis->FastCoorToLogical(DataY);

		/* Output point or line */
    if (ValidLastPoint) {
      pDC->LineTo(LogX, LogY);
     }
    else {
      pDC->MoveTo(LogX, LogY);
      pDC->SetPixel(LogX, LogY, m_DataColor);
      ValidLastPoint = true;
     }

		/* Draw a data point if required */
    if (m_DisplayPoints) {
      pDC->FillSolidRect(LogX-1, LogY-1, 3, 3, 0);
      pDC->MoveTo(LogX, LogY);
     }

   }

  return (0);
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::DrawData()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void DrawPoint (X, Y, pDC);
 *
 * Outputs a single point to the graph.
 *
 *=========================================================================*/
void CGenGraphData::DrawPoint (const float X, const float Y, CDC* pDC) {
  if (!HasValidAxes()) return;	/* Ignore if graph doesn't have valid axes */

	/* See if we're scanning forwards or backwards */
  m_ScanForwards = (X >= m_ScanX);

	/* Clear the current scan line, if any */
  ClearScan(X, pDC);

	/* Check for auto scrolling conditions to update X-Axis */
  AutoScrollXAxis(X, pDC);

	/* Output scan point to graph */
  DrawScan(X, pDC);
  DrawPointPriv (X, Y, pDC);

	/* Store the point for the display of the next data point */
  AddBufferPoint(X, Y);	

	/* Update members for next point output */
  m_HasLastPoint = true;
  m_LastX = X;
  m_LastY = Y;
  m_LastScanForwards = m_ScanForwards;
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::DrawPoint()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void DrawPointPriv (X, Y, pDC);
 *
 * Protected class method to output a single scan point.
 *
 *=========================================================================*/
void CGenGraphData::DrawPointPriv (const float X, const float Y, CDC* pDC) {
  //DEFINE_FUNCTION("CGenGraphData::DrawPointPriv()");
  int LogX;
  int LogY;

	/* Compute the point's X/Y location */
  LogX = m_pXAxis->LimitCoorToLogical(X);
  LogY = m_pYAxis->LimitCoorToLogical(Y);
 
	/* Determine whether to draw a line or simply a point */
  if (m_HasLastPoint) {
    pDC->SelectObject(m_pDataPen);
    pDC->MoveTo(m_pXAxis->LimitCoorToLogical(m_LastX), m_pYAxis->LimitCoorToLogical(m_LastY));
    pDC->LineTo(LogX, LogY);
   }
  else {
    pDC->SetPixel(LogX, LogY, m_DataColor);
   }

 }
/*===========================================================================
 *		End of Class Method CGenGraphData::DrawPointPriv()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void DrawScan (const float X, CDC* pDC);
 *
 * Outputs the scan line at the given position in the graph.
 *
 *=========================================================================*/
void CGenGraphData::DrawScan (const float X, CDC* pDC) {
  int LogX;

	/* Ignore if required */
  if (!m_DisplayScan) return;

  if (m_pXAxis->IsValidPoint(X)) {
    LogX = GetScanLocation(X, m_ScanForwards);
    m_ScanX = X;

    pDC->SelectObject(m_pScanPen);
    pDC->MoveTo (LogX, m_pYAxis->GetTop() + 1);
    pDC->LineTo (LogX, m_pYAxis->GetBottom());
   }

 }
/*===========================================================================
 *		End of Class Method CGenGraphData::DrawScan()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - int GetScanLocation (X, ScanForwards);
 *
 * Converts the given coordinate into a logical position for the scan line.
 * Protected class method.
 *
 *=========================================================================*/
int CGenGraphData::GetScanLocation (const float X, const bool ScanForwards) {
  int LogX;

	/* Compute the logical X coordinate */
  if (ScanForwards)
    LogX = m_pXAxis->LimitCoorToLogical(X) + 1;
  else
    LogX = m_pXAxis->LimitCoorToLogical(X) - 1;

	/* Ensure we don't exceed the graph control size */
  if (LogX >= m_pXAxis->GetRight())
    LogX = m_pXAxis->GetRight() - 1;
  else if (LogX <= m_pXAxis->GetLeft())
    LogX = m_pXAxis->GetLeft() + 1;

  return (LogX);
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::GetScanLocation()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Event - int OnPaint (pDC);
 *
 *=========================================================================*/
int CGenGraphData::OnPaint (CDC* pDC) {
  
  if (IsVisible()) { 
    DrawBG(pDC);
    DrawBuffer(pDC);
    DrawData(pDC);
   }

  return (0);
 }
/*===========================================================================
 *		End of Class Event CGenGraphData::OnPaint()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void ResetScan (void);
 *
 * Resets all the scan parameters.
 *
 *=========================================================================*/
void CGenGraphData::ResetScan (void) {
  float Range;

  m_HasLastPoint = false;
  m_ScanX = 0;

  if (m_pXAxis != NULL) {
    Range = m_pXAxis->GetEndCoor() - m_pXAxis->GetStartCoor();
    m_pXAxis->SetStartCoor(0);
    m_pXAxis->SetEndCoor(Range);
   }

 }
/*===========================================================================
 *		End of Class Method CGenGraphData::ResetScan()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void SetAxes (pXAxis, pYAxis);
 *
 *=========================================================================*/
void CGenGraphData::SetAxes (CGenGraphAxis* pXAxis, CGenGraphAxis* pYAxis) {
  DEFINE_FUNCTION("CGenGraphData::SetAxes()");

	/* Ensure valid input */
  ASSERT(pXAxis != NULL && pYAxis != NULL);
  ASSERT(pXAxis->IsXAxis() && pYAxis->IsYAxis());

  m_pXAxis = pXAxis;
  m_pYAxis = pYAxis;
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::SetAxes()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void SetFloatData (pXData, pYData, NumPoints);
 *
 *=========================================================================*/
void CGenGraphData::SetFloatData (float* pXData, float* pYData, const int NumPoints) {
  DestroyData();

  if (pXData != NULL) {
    m_XData.m_pData = pXData;
    m_XData.m_Referenced = true;
    m_XData.m_NumPoints = NumPoints;
    m_XData.m_SkipBytes = 0;
   }

  m_YData.m_pData = pYData;
  m_YData.m_Referenced = true;
  m_YData.m_NumPoints = NumPoints;
  m_YData.m_SkipBytes = 0;
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::SetFloatData()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void SetFloatDataEx (pXData, XSkipBytes, pYData, YSkipBytes, NumPoints);
 *
 *=========================================================================*/
void CGenGraphData::SetFloatDataEx (float* pXData, const int XSkipBytes, 
				    float* pYData, const int YSkipBytes, 
				    const int NumPoints) {
  DestroyData();

  if (pXData != NULL) {
    m_XData.m_pData = pXData;
    m_XData.m_Referenced = true;
    m_XData.m_NumPoints  = NumPoints;
    m_XData.m_SkipBytes  = XSkipBytes;
   }

  m_YData.m_pData = pYData;
  m_YData.m_Referenced = true;
  m_YData.m_NumPoints  = NumPoints;
  m_YData.m_SkipBytes  = YSkipBytes;
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::SetFloatDataEx()
 *=========================================================================*/


/*===========================================================================
 *
 * Class CGenGraphData Method - void SetLineColor (Color);
 *
 *=========================================================================*/
void CGenGraphData::SetLineColor (const COLORREF Color) {
  DEFINE_FUNCTION("CGenGraphData::SetLineColor()");

  DestroyPointer(m_pDataPen);
  m_DataColor = Color;
  m_pDataPen  = new CPen(PS_SOLID, 1, m_DataColor);
  ValidateNewPointer(m_pDataPen);
 }
/*===========================================================================
 *		End of Class Method CGenGraphData::SetLineColor()
 *=========================================================================*/
